@@ -14,6 +14,7 @@ TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR TH
 import os, sys, time, json
 import pandas as pd
 from copy import deepcopy
+from io import StringIO
 import pickle
 
 from .constants import *
@@ -42,13 +43,15 @@ class BaseRunner:
         """
         return str(uuid.uuid4())
 
-    def _buffer_save(self, df, batch_size=1000, client_id=None, request_id=None, verbose=False):
+    def _buffer_save(self, df, batch_size=1000, client_id=None, request_id=None, verbose=False, compressed=False, staging=False):
         """
         :param df:
         :param batch_size:
         :param client_id:
         :param request_id:
         :param verbose:
+        :param compressed:
+        :param staging:
         :return res:
         """
         if verbose: print('Saving data to buffer...')
@@ -61,7 +64,7 @@ class BaseRunner:
         for batch in batched_df:
             idx += 1
             for attempt in range(0, 5):
-                if self._batch_save(batch, idx, len(batched_df), client_id, request_id, verbose=(attempt==4)):
+                if self._batch_save(batch, idx, len(batched_df), client_id, request_id, verbose=(attempt==4), compressed=compressed, staging=staging):
                     if verbose: sys.stdout.write('        Processed batch {} of {}\r'.format(idx, len(batched_df)))
                     success += 1
                     break
@@ -70,25 +73,35 @@ class BaseRunner:
         if verbose: sys.stdout.write('\n')
         return {'request_id': request_id, 'total_batches': len(batched_df), 'batches_saved': success}
 
-    def _batch_save(self, batch, idx, n, client_id, request_id, verbose=False):
+    def _batch_save(self, batch, idx, n, client_id, request_id, verbose=False, compressed=False, staging=False):
         """
+        If staging is false, converts dataframe to dictionary and uploads to database buffer.
+        If staging is true, converts dataframe to CSV chunk and uploads to data lake buffer.
+
         :param batch:
         :param idx:
         :param n:
         :param client_id:
         :param request_id:
         :param verbose:
+        :param compressed:
         :return success:
         """
         success = False
         uri = '{}/buffer/'.format(self._base_url)
+        # if DEBUG: print('[_batch_save] batch #{} len: {} staging: {}'.format(idx, len(batch), staging))
         try:
-            res = self._client._post(uri, {
-                'command': 'upload',
-                'data': batch.to_dict(),
-                'request_id': request_id,
-                'client_id': client_id,
-            })
+            res = self._client._post(
+                uri,
+                {
+                    'command': 'upload',
+                    'data': batch.to_dict() if not staging else batch.to_csv(header=(idx==1), index=False),
+                    'request_id': request_id,
+                    'client_id': client_id,
+                    'staging': staging,
+                },
+                compressed=compressed,
+            )
             if res['status']==200:
                 success = True
             else:
@@ -97,29 +110,32 @@ class BaseRunner:
             if verbose: print('[__batch_save] WARNING! API call failed for batch {} of {}'.format(idx, n))
         return success
 
-    def _buffer_read(self, client_id=None, request_id=None, dataframe_name='df', verbose=False, raw=False):
+    def _buffer_read(self, client_id=None, request_id=None, dataframe_name='df', verbose=False, raw=False, staging=False):
         """
         :param client_id:
         :param request_id:
         :param verbose:
         :param raw:
+        :param staging:
         :return df:
         """
-        if DEBUG: print('[_buffer_read] making call to read buffer')
-        res2 = self.__read(client_id=client_id, request_id=request_id, dataframe_name=dataframe_name, verbose=verbose)
-        if DEBUG: print('[_buffer_read] Buffer returned response')
+        # if DEBUG: print('[_buffer_read] making call to read buffer')
+        res2 = self.__read(client_id=client_id, request_id=request_id, dataframe_name=dataframe_name, verbose=verbose, staging=staging)
+        # if DEBUG: print('[_buffer_read] Buffer returned response')
         if raw: return res2
         if res2['status']!=200:
             print('ERROR! Buffer read failed: {}'.format(res2))
             return pd.DataFrame(None)
-        if DEBUG: print('[_buffer_read] Buffer read successfully')
-        return pd.DataFrame(res2['response']['data'])
+        # if DEBUG: print('[_buffer_read] Buffer read successfully')
+        if res2['response']['data'] is None: return pd.DataFrame()
+        return pd.DataFrame(res2['response']['data']) if not staging else pd.read_csv(StringIO(res2['response']['data']))
 
-    def __read(self, client_id=None, request_id=None, dataframe_name='df', verbose=False):
+    def __read(self, client_id=None, request_id=None, dataframe_name='df', verbose=False, staging=False):
         """
         :param client_id:
         :param request_id:
         :param verbose:
+        :param staging:
         :return res:
         """
         uri = '{}/buffer/'.format(self._base_url)
@@ -128,13 +144,60 @@ class BaseRunner:
             'request_id': request_id,
             'client_id': client_id,
             'dataframe_name': dataframe_name,
+            'staging': staging,
         })
 
+    def queue_purge(self, client_id=None, verbose=False):
+        """
+        :param client_id:
+        :param verbose:
+        :return res:
+        """
+        if verbose: print('Purging queue...')
+        uri = '{}/analytics/'.format(self._base_url)
+        res = self._client._post(uri, {
+            'command': 'purge',
+            'client_id': client_id,
+        })
+        if res['status']!=200: print('WARNING! Queue purge failed: {}'.format(res))
+        return res
+
+    def buffer_purge(self, client_id=None, verbose=False):
+        """
+        :param client_id:
+        :param verbose:
+        :return res:
+        """
+        if verbose: print('Purging buffer...')
+        uri = '{}/buffer/'.format(self._base_url)
+        res = self._client._post(uri, {
+            'command': 'purge',
+            'client_id': client_id,
+        })
+        if res['status']!=200: print('WARNING! Buffer purge failed: {}'.format(res))
+        return res
+
+    def buffer_usage(self, client_id=None, verbose=False):
+        """
+        :param client_id:
+        :param verbose:
+        :return res:
+        """
+        if verbose: print('Checking buffer usage...')
+        uri = '{}/buffer/'.format(self._base_url)
+        res = self._client._post(uri, {
+            'command': 'usage',
+            'client_id': client_id,
+        })
+        if res['status']!=200: print('WARNING! Checking buffer usage failed: {}'.format(res))
+        return res
+
     def _buffer_clear(self, client_id=None, request_id=None, verbose=False):
         """
         :param client_id:
         :param request_id:
         :param verbose:
+        :param staging:
         :return res:
         """
         if verbose: print('Clearing buffer...')
@@ -364,13 +427,37 @@ class BaseRunner:
             try:
                 res = self._client._post(uri, payload)
                 if verbose: sys.stdout.write('[_poll][{}] {}\r'.format(counter, res))
-                if res['response']['status']=='Complete': break
+                if res['response']['status'] in ['Complete', 'Failed'] or 'Failed:' in res['response']['status']: break
             except:
                 pass
             counter += step
         if verbose: sys.stdout.write('\n')
         return res
 
+    def _status(self, request_id=None, client_id=None, verbose=False):
+        """
+        :param request_id:
+        :param client_id:
+        :param verbose:
+        :return res:
+        """
+        res = {}
+        uri = '{}/analytics/'.format(self._base_url)
+        if request_id is None or client_id is None:
+            print('[_status] ERROR! Invalid status request (request_id: {}, client_id: {})'.format(request_id, client_id))
+        else:
+            try:
+                res = self._client._post(uri, {'request_id': request_id, 'client_id': client_id, 'command': 'task-status'})
+            except:
+                if verbose: print('[_status] ERROR! Could not retrieve status for request ID: {} and client ID: {}'.format(request_id, client_id))
+            else:
+                if res['status']==200:
+                    res = res['response']
+                else:
+                    if verbose: print('[_status] ERROR! Status inquiry for request ID: {} and client ID: {} returned status: {}'.format(request_id, client_id, res['status']))
+                    res = {}
+        return res
+
     def _keys_save(self, model_id=None, keys=None, verbose=False):
         """
         :param model_id:
